
# Real-time Processor Architectures for Worst Case Execution Time Reduction
# Thesis Software Distribution.
# This code is licensed under GPL v2, and is copyright (C) Jack Whitham 2004-08.
#


import os, math
import collections


# Solver command line, must include %s for input file name substitution.
# Output must be sent to stdout.
# Omit the path to use your system's version of glpsol.
GLPSOLVE_CMD_LINE = 'glpsol --tmlim 500 --math %s'

[ VAR_BINARY, VAR_INTEGER, VAR_MIXED ] = range(2, 5)


class ILPError(Exception): 
    pass
class ILPTimeoutError(ILPError): 
    pass

class ILP_Solver_GLPK:
    def __init__(self, model_file_name):
        self.model_file_name = model_file_name + '.mod'
        self.result_key = '!result: '
        self.Reset_All()

    def Reset_All(self):
        self.sect0 = []
        self.sect1 = collections.deque() # [] does not have
        self.sect2 = collections.deque() # the .clear() method
        self.sect3 = []
        self.sect4 = []
        self.all_sectors = [ [ '/* generated */' ],
                self.sect0, self.sect1,
                self.sect2, [ 'solve ;' ], self.sect3, 
                [ 'data ;' ], self.sect4, [ 'end ;' ] ]

    def Add_Parameter(self, pname, pval):
        self.sect0.append('param %s, integer, >= 0 ;\n' % pname)
        self.sect4.append('param %s := %u ;\n' % (pname, pval))

    def Add_Comment(self, comment):
        assert type(comment) == str
        self.sect4.append('\n/*')
        self.sect4.append(comment)
        self.sect4.append('*/\n')

    def Add_Var(self, vname, var_type):
        printf_code = '%d'
        if ( var_type == VAR_INTEGER ):
            self.sect0.append('var %s, integer, >= 0 ;\n' % vname)
        elif ( var_type == VAR_BINARY ):
            self.sect0.append('var %s, binary ;\n' % vname)
        elif ( var_type == VAR_MIXED ):
            self.sect0.append('var %s, >= 0 ;\n' % vname)
            printf_code = '%1.0f'
        else:
            assert False, "Unknown var_type %s" % var_type

        self.sect3.append(r'printf "%s%s = %s\n", %s;' %
                    (self.result_key, vname, printf_code, vname))
        self.sect3.append('\n')

    def Reset_Terms(self):
        self.sect1.clear()
        self.sect2.clear()

    def Add_Term(self, lhs, operator, rhs):
        assert not '(' in lhs
        assert not '(' in rhs
        self.sect1.append('s.t. c%05u: %s %s %s ;\n' %
                    (len(self.sect1), lhs, operator, rhs))

    def Add_User_Terms(self, user):
        self.sect1.append('\n/* user constraints start */\n')
        for t in user:
            self.sect1.append('s.t. u%05u: %s\n' % (len(self.sect1), t))
        self.sect1.append('\n/* user constraints end */\n')

    def Add_Problem(self, maximise_this):
        self.sect2.append('maximize obj: %s ;\n' % maximise_this)

    def Solve(self, receive_fn):
        model_file_name = self.model_file_name
        old_name = model_file_name + '.old'

        try:
            os.unlink(old_name)
        except:
            pass
        try:
            os.rename(model_file_name, old_name)
        except:
            pass

        model_file = file(model_file_name, 'wt')
        for sect in self.all_sectors:
            model_file.write(''.join(sect))
            model_file.write('\n\n')
        model_file.close()

        return self.Call_Solver(model_file_name, receive_fn)

    def Call_Solver(self, model_file_name, receive_fn):


        cmdline = GLPSOLVE_CMD_LINE % model_file_name

        try:
            glpsol = os.popen(cmdline)
        except:
            raise ILPError('Could not launch ILP solver: %s' % cmdline)

        nonzero = False
        timeout = False
        all = []
        errors = []
        commit = []
        for line in glpsol:
            if ( line.find('TIME LIMIT EXCEEDED') >= 0 ):
                timeout = True
                
            elif (( line.find('PROBLEM') >= 0 )
            or ( line.find('error') >= 0 )):
                errors.append(line)

            all.append(line)

            if ( line.startswith(self.result_key) ):
                params = line[ len(self.result_key) : ]
                words = params.split('=')
                assert len(words) == 2
                key = words[ 0 ].strip()
                value = int(words[ 1 ])
                assert value >= 0

                commit.append((key, value))

                if ( value > 0 ):
                    nonzero = True

        glpsol.close()

        if ( timeout ):
            raise ILPTimeoutError()

        elif ( nonzero ):
            # Commit changes
            for (key, value) in commit:
                receive_fn(key, value)

        else:
            if ( len(errors) == 0 ):
                errors.append('No errors?')
                errors.append('** '.join(all[ : min(len(all), 10) ]))

            # Error - all results were zero
            raise ILPError('Could not solve: %s\n%s' % (cmdline,
                '\n'.join(errors)))

        return ''.join(all)

class ILP_Solver_LPS(ILP_Solver_GLPK):
    def __init__(self, model_file_name):
        # For debugging - simultaneously generate a model for another solver :)
        #self.alternate = None
        self.alternate = ILP_Solver_GLPK(model_file_name + '2')


        ILP_Solver_GLPK.__init__(self, model_file_name)

        import lpsolve55 

        self.lpsolve_module = lpsolve55
        self.lp_fn = self.lpsolve_module.lpsolve


    def Reset_All(self):
        if ( self.alternate != None ):
            self.alternate.Reset_All()
        
        # objective function
        self.sect_obj = collections.deque()
        # constraints
        self.sect_constraints = collections.deque()
        self.sect_var_constraints = []
        # variable decls
        self.sect_vars = []
        self.variables = set()

        self.all_sectors = [ ['/* lp_solve model */'],
                self.sect_obj, self.sect_constraints,
                self.sect_var_constraints,
                self.sect_vars ]


    def Add_Parameter(self, pname, pval):
        if ( self.alternate != None ):
            self.alternate.Add_Parameter(pname, pval)

        self.sect_vars.append('int %s ;\n' % pname)
        self.sect_var_constraints.append('%s = %u ;\n' % (pname, pval))

    def Add_Comment(self, comment):
        assert type(comment) == str
        self.sect_vars.append('\n/*')
        self.sect_vars.append(comment)
        self.sect_vars.append('*/\n')

    def Add_Var(self, vname, var_type):
        if ( self.alternate != None ):
            self.alternate.Add_Var(vname, var_type)

        self.sect_vars.append('int %s ;\n' % vname)
        if ( var_type in (VAR_MIXED, VAR_INTEGER) ):
            # is non-negative integer
            self.sect_var_constraints.append('0 <= %s ;\n' % vname)
        elif ( var_type == VAR_BINARY ):
            # is binary
            self.sect_var_constraints.append('0 <= %s <= 1 ;\n' % vname)
        else:
            assert False, "Unknown var_type %s" % var_type
        self.variables.add(vname)

    def Reset_Terms(self):
        if ( self.alternate != None ):
            self.alternate.Reset_Terms()

        self.sect_constraints.clear()
        self.sect_obj.clear()

    def Add_Term(self, lhs, operator, rhs):
        if ( self.alternate != None ):
            self.alternate.Add_Term(lhs, operator, rhs)

        assert not '(' in lhs
        assert not '(' in rhs


        self.sect_constraints.append('%s %s %s ;\n' % (lhs, operator, rhs))

    def Add_User_Terms(self, user):
        if ( self.alternate != None ):
            self.alternate.Add_User_Terms(user)

        for t in user:
            assert not t.startswith('s.t.'), (
                "Your constraints file seems to be in GLPK format. " )
            self.sect_constraints.append('%s\n' % t)

    def Add_Problem(self, maximise_this):
        if ( self.alternate != None ):
            self.alternate.Add_Problem(maximise_this)

        self.sect_obj.append('max: %s ;\n' % maximise_this)

    def Call_Solver(self, model_file_name, receive_fn):

        if ( self.alternate != None ):
            def Dummy(*p):
                pass
            try:
                self.alternate.Solve(Dummy)
                print 'ALTERNATE SOLVER OK'
            except Exception, x:
                print 'ALTERNATE SOLVER CRASHED: %s' % str(x)


        lpmod = self.lpsolve_module
        lp = self.lp_fn('read_lp', model_file_name)
        self.lp_fn('set_verbose', lp, lpmod.NORMAL)
        self.lp_fn('set_presolve', lp, 
                    lpmod.PRESOLVE_ROWS | lpmod.PRESOLVE_MERGEROWS )
        self.lp_fn('solve', lp)

        # The same protocol is used in lp_report.c to retrieve results,
        # but there, precision is destroyed by printf "%12g". 
        
        orig_columns = self.lp_fn('get_Norig_columns', lp)
        orig_rows = self.lp_fn('get_Norig_rows', lp)
        assert orig_columns > 0
        assert orig_rows > 0

        MAX_EPSILON = 0.25

        for col in xrange(1, orig_columns + 1):
            vname = self.lp_fn('get_origcol_name', lp, col)
            value_f = self.lp_fn('get_var_primalresult',  lp, 
                        orig_rows + col)
            assert type(vname) == str
            assert value_f > -MAX_EPSILON, "Value is < 0: %s = %1.5f" % (
                        vname, value_f)

            value_i = math.floor(value_f + 0.5)
            epsilon = abs(value_f - value_i)
            assert epsilon < MAX_EPSILON, (
    "lp_solve did not return an integer value for variable %s: %1.5f" % (
            vname, value_f))

            value = int(value_i)

            receive_fn(vname, value)
                

        self.lp_fn('delete_lp', lp)

        return ''

def Read_User_Constraints(fname):
    user_constraints = []
    for line in file(fname):
        line = line.strip()
        if ( len(line) > 0 ):
            user_constraints.append(line)
    return user_constraints


ILP_Solver = ILP_Solver_GLPK




